/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef MONDATAMSG_HH
#define MONDATAMSG_HH

#include "lmsg/Message.hh"
#include "TSeries.hh"
#include "FSeries.hh"
#include "FSpectrum.hh"
#include "FSeries.hh"
#include "Histogram1.hh"
#include "fSeries/DFT.hh"
#include "fSeries/PSD.hh"
#include "fSeries/ASD.hh"
#include <iosfwd>

namespace lmsg {
    class TransInput;
    class TransOutput;
}

//-------------------------------------  Enumerate monitor data message types
/** Enumerate the monitor data messages.
 */
enum {
    mdm_Null=65536,  ///< Null message. Set the start of the monitor messages
    mdm_Int,         ///< Integer data word message
    mdm_Double,      ///< Double precision data word message
    mdm_String,      ///< Character string data message.
    mdm_TSeries,     ///< %Time series object message
    mdm_FSeries,     ///< Frequency series (DFT) object message
    mdm_FSpectrum,   ///< Spectrum (ASD, PSD) object message
    mdm_Request,     ///< Request a data object
    mdm_Index,       ///< Index of available data objects
    mdm_ReqIndex,    ///< request an index of data objects
    mdm_NAck,        ///< Negative acknowlege (request failed) message
    mdm_Histogram1   ///< 1D Histogram data object message
};

/** The monitor data message template is used to construct all of the monitor 
 *  data message structures. Each data message contains the data object name 
 *  and the data object. %TransInput and %TransOutput (streamer) instances are
 *  used to transfer the serialized data to and from the object.
 */
template<class T, unsigned int MsgID>
class MonDataMsg : public lmsg::Message {
public:
    /** Construct and empty data message.
     */
    MonDataMsg(void);
    
    /** Destroy the message
     *  \brief destructor.
     */
    ~MonDataMsg(void);

    /** Construct a data message from a data object.
     *  \brief Construct an initialized data message.
     *  \param name Data object name
     *  \param ival Object from which the message data is initialized. 
     */
    MonDataMsg(const std::string& name, const T& ival);

    /** Construct a data mesage from a stream.
     *  \brief stream constructor
     *  \param istr Stream from which the data message is initialized.
     */
    MonDataMsg(lmsg::TransInput& istr);

    /** Set the message structure from a data stream.
     *  \brief Set the data
     *  \param istr Serialized data stream.
     */
    void setData(lmsg::TransInput& istr);

    /** Copy the message data to a serialized stream.
     *  \brief Get the data
     *  \param ostr Serialized data stream.
     */
    void getData(lmsg::TransOutput& ostr) const;

    /** Get the pyload data object content.
     *  \brief Get the pyload data
     *  \return Data object content.
     */
    T getValue(void) const;

    /** Get the requested object name string.
     *  \brief Requested object name
     *  \returns character string pointer.
     */
    const char* getName(void) const;

    /** Get the message structure length.
     *  \brief Message structure length
     *  \returns Length of message data.
     */
    lmsg::size_type size(void) const;

    /** Get the message type.
     *  \brief message type
     *  \returns message type
     */
    lmsg::MsgHeader::MsgType getType(void) const;

private:
    std::string mName;
    T mValue;
};

//-------------------------------------  Integer message class
typedef MonDataMsg<long,        mdm_Int>       MDM_Int;
typedef MonDataMsg<double,      mdm_Double>    MDM_Double;
typedef MonDataMsg<std::string, mdm_String>    MDM_String;
typedef MonDataMsg<TSeries,     mdm_TSeries>   MDM_TSeries;
typedef MonDataMsg<FSeries,     mdm_FSeries>   MDM_FSeries;
typedef MonDataMsg<FSpectrum,   mdm_FSpectrum> MDM_FSpectrum;
typedef MonDataMsg<Histogram1,  mdm_Histogram1> MDM_Histogram1;  // Histogram1


//-------------------------------------  fSeries: duplicate message ids
typedef MonDataMsg<containers::DFT, mdm_FSeries>   MDM_DFT;
typedef MonDataMsg<containers::ASD, mdm_FSpectrum> MDM_ASD;
typedef MonDataMsg<containers::PSD, mdm_FSpectrum> MDM_PSD;

/** Data request message structure.
 *  \brief Data request message structure.
 */
class MDM_Request : public lmsg::Message {
public:
    /** Default constructor.
     */
    MDM_Request(void);

    /** Destroy a request message.
     *  \brief destructor.
     */
    ~MDM_Request(void);

    /** Construct a request for a named data object.
     *  \brief Data request message.
     *  \param name Name of requested object.
     */
    MDM_Request(const std::string& name);

    /** Construct a request from a serialised data string.
     *  \brief Construct a data request.
     *  \param istr Stream containing serialized request info.
     */
    MDM_Request(std::istream& istr);

    /** Get the requested object name string.
     *  \brief Requested object name
     *  \returns character string pointer.
     */
    const char* getName(void) const;

    /** Get the message structure length.
     *  \brief Message structure length
     *  \returns Length of message data.
     */
    lmsg::size_type size(void) const;

    /** Get the message type.
     *  \brief message type
     *  \returns message type
     */
    lmsg::MsgHeader::MsgType getType(void) const;

    /** Set the message structure from a data stream.
     *  \brief Set the data
     *  \param istr Serialized data stream.
     */
    void setData(lmsg::TransInput& istr);

    /** Copy the message data to a serialized stream.
     *  \brief Get the data
     *  \param ostr Serialized data stream.
     */
    void getData(lmsg::TransOutput& ostr) const;

private:
    std::string mName;
};

//-------------------------------------  Data request message
class MDM_Index : public lmsg::Message {
public:
    /** Default monitor index message constructor.
     *  \brief default constructor.
     */
    MDM_Index(void);
    /** Destroy a monitor index message.
     *  \brief Destructor.
     */
    ~MDM_Index(void);
    /** Construct a monitor index message from a string.
     *  \brief string constructor.
     *  \param index Index message contents.
     */ 
    MDM_Index(const std::string& index);
    /** Construct an index data message from a input stream.
     *  \brief Stream constructor.
     *  \param istr Input stream.
     */
    MDM_Index(std::istream& istr);
    
    /** Get the index data string.
     *  \brief get the index
     *  \param s String into which the index is copied.
     */
    void getIndex(std::string& s) const;

    /** Get the message data length.
     *  \brief Message data length
     *  \returns Length of message data.
     */
    lmsg::size_type size(void) const;

    /** Get the message type.
     *  \brief message type
     *  \returns message type
     */
    lmsg::MsgHeader::MsgType getType(void) const;

    /** Set the message structure from a data stream.
     *  \brief Set the data
     *  \param istr Serialized data stream.
     */
    void setData(lmsg::TransInput& istr);

    /** Copy the message data to a serialized stream.
     *  \brief Get the data
     *  \param ostr Serialized data stream.
     */
    void getData(lmsg::TransOutput& ostr) const;

private:
    std::string mIndex;
};

/** Index request message consist of a header only.
 */
typedef lmsg::HdrOnlyMsg<mdm_ReqIndex> MDM_ReqIndex;

/** Monitor negative acknowlege (NAck) message consist of a header only.
 */
typedef lmsg::HdrOnlyMsg<mdm_NAck>     MDM_NAck;

#endif  //  MONDATAMSG_HH
