/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef MONSERVER_HH
#define MONSERVER_HH

#ifndef DMTOFFLINE
#include "AppServer.hh"
#endif
#include <map>
#include <string>
#include <stdexcept>

class DataDesc;
class TSeries;
class FSeries;
class FSpectrum;
class Histogram1;
namespace containers {
    class ASD;
    class DFT;
    class PSD;
}


/**  The monitor data server API allows the user to specify data elements 
  *  to be served on demand to clients. The monitor server maintains a 
  *  list of the data elements and their names.
  *  @memo Monitor data server.
  *  @author John Zweizig
  *  @version 1.1; Modified April 20, 2000
  */
class MonServer 
#ifndef DMTOFFLINE
: private lmsg::AppServer 
#endif
{
public:
  typedef std::map<std::string, DataDesc> Dictionary;
  typedef Dictionary::iterator dict_iter;
  typedef Dictionary::const_iterator const_dict_iter;

public:
  /**  Construct a monitor data base and register the server with the 
    *  name server.
    */
  MonServer(const std::string& Process="");

  /**  Close the server port and unregister its name.
    *  \brief Close the server.
    */
  virtual ~MonServer(void);

  /**  Test whether data have been defined for service.
   */
  bool dataServed(void) const;

  /**  Process a message from the monitor input port.
   */
  void processMsg(void);

  /**  Handle the attention interrupt delivered to a monitor.
    *  @memo Monitor attention handler.
    */
  void Attention(void);

  /** Specify a monitor data object to be served on demand. Series data 
   *  are converted to single-precision float values before being sent 
   *  to the client. Objects implemented for serve_data include TSeries, 
   *  FSeries, FSpectra, Histogram1, containers::ASD, containers::DFT 
   *  and containers::PSD.
   *  \brief Serve a power spectrum.
   *  \param Name Data name
   *  \param fs   Pointer to data to be served.
   *  \param Comm Comment supplied with index.
   */
  template<class T> 
  void serveData(const std::string& Name, const T* iVal,
		 const std::string& Comm="");

  /**  Set the server name.
   */
  void setServerName(const std::string& Name);

  /**  Get the debug flag valus
   */
  int getDebug(void) const;

  /**  Set the debug flag
   */
  void setDebug(int lvl);

  /**  Return a reference to the served dictionary data.
    */
  const Dictionary& refDictionary(void) const;

private:
  Dictionary mDict;
#ifdef DMTOFFLINE
  int  mDebug;
#endif
};

//======================================  Data Descriptor class
/** The data descriptor class contains a description of a data object to 
 *  be served by the monitor server interface. %DataDesc instances are
 *  kept in the monitor service dictionary.
 */
class DataDesc {
public:
    /** Numerate all the data types that may be served.
     */
    enum DataType {
	t_void,          ///< not a data object
	t_long,          ///< long integer data word
	t_double,        ///< double precision float data word
	t_string,        ///< character string data object
	t_TSeries,       ///< time series data object
	t_FSeries,       ///< obsolete dft data object
	t_FSpectrum,     ///< obsolete asd/psd data object
	t_Histogram1,    ///< 1-D histogram data object 
	t_asd,           ///< asd data pbject
	t_dft,           ///< dft data object
	t_psd            ///< psd data object
    };

    /** construct an empty data descriptor.
     */
    DataDesc(void);

    /** Construct a long integer data word descriptor.
     *  \brief Long integer descriptor.
     *  \param addr Constant pointer to the data word.
     *  \param Comm Optional comment.
     */
    DataDesc(const long*        addr,     const std::string& Comm="");

    /** Construct a double precieion float data word descriptor.
     *  \brief Double float descriptor.
     *  \param addr Constant pointer to the data word.
     *  \param Comm Optional comment.
     */
    DataDesc(const double*      addr,     const std::string& Comm="");

    /** Construct a character string object descriptor.
     *  \brief Character string data descriptor.
     *  \param addr Constant pointer to the data.
     *  \param Comm Optional comment.
     */
    DataDesc(const std::string* addr,     const std::string& Comm="");

    /** Construct a time series data object descriptor.
     *  \brief Time series data descriptor.
     *  \param addr Constant pointer to the data.
     *  \param Comm Optional comment.
     */
    DataDesc(const TSeries*     addr,     const std::string& Comm="");

    /** Construct an %FSeries (obsolete) data descriptor.
     *  \brief %FSeries data descriptor.
     *  \param addr Constant pointer to the data.
     *  \param Comm Optional comment.
     */
    DataDesc(const FSeries*     addr,     const std::string& Comm="");

    /** Construct %FSpectrum (obsolete) data descriptor.
     *  \brief Long integer descriptor.
     *  \param addr Constant pointer to the data.
     *  \param Comm Optional comment.
     */
    DataDesc(const FSpectrum*   addr,     const std::string& Comm="");

    /** Construct a 1D histogram data descriptor.
     *  \brief 1D Histogram descriptor.
     *  \param addr Constant pointer to the data.
     *  \param Comm Optional comment.
     */
    DataDesc(const Histogram1*  addr,     const std::string& Comm="");

    /** Construct an ASD data descriptor.
     *  \brief ASD data object descriptor.
     *  \param addr Constant pointer to the data word.
     *  \param Comm Optional comment.
     */
    DataDesc(const containers::ASD* addr, const std::string& Comm="");

    /** Construct a DFT data descriptor.
     *  \brief DFT data object descriptor.
     *  \param addr Constant pointer to the data word.
     *  \param Comm Optional comment.
     */
    DataDesc(const containers::DFT* addr, const std::string& Comm="");

    /** Construct a PSD data descriptor.
     *  \brief PSD data object descriptor.
     *  \param addr Constant pointer to the data word.
     *  \param Comm Optional comment.
     */
    DataDesc(const containers::PSD* addr, const std::string& Comm="");

    /** Destroy a data descriptor.
     *  \brief Destructor.
     */
    ~DataDesc(void);

    /** Get a constant pointer to the data object.
     *  \brief get data address.
     *  \returns constant pointer to the described data object.
     */
    const void* getAddr() const;

    /** Get a constant pointer to the null-terminated comment string.
     *  \brief Get the comment string
     *  \return pointer to the null-terminated comment string.
     */
    const char* getComment() const;

    /** Get the data object type.
     *  \brief data type
     *  \return data type enumerator value.
     */
    DataType getType() const;
    
private:
  DataType    mType;
  const void* mAddr;
  std::string mComment;
#ifdef DMTOFFLINE
  int         mDebug;
#endif
};

//======================================  MonServer inline functions.
template<class T>
inline void
MonServer::serveData(const std::string& Name, const T* data,
		     const std::string& Comm) {
    if (!data) throw std::runtime_error("serveData: Invalid data address.");
    if (Name.empty()) throw std::runtime_error("serveData: Invalid name string.");
    mDict[Name] = DataDesc(data, Comm);
}

inline bool
MonServer::dataServed(void) const {
    return !mDict.empty();
}

inline const MonServer::Dictionary&
MonServer::refDictionary(void) const {
    return mDict;
}

#ifdef DMTOFFLINE
inline void 
MonServer::setDebug(int lvl) {
    mDebug = lvl;
}

inline int  
MonServer::getDebug(void) const {
    return mDebug;
}
#else
inline void 
MonServer::setDebug(int lvl) {
    AppServer::setDebug(lvl);
}

inline int  
MonServer::getDebug(void) const {
    return AppServer::getDebug();
}
#endif

//======================================  MonServer inline functions.
inline const void* 
DataDesc::getAddr(void) const {
    return mAddr;
}

inline const char* 
DataDesc::getComment(void) const {
    return mComment.c_str();
}

inline DataDesc::DataType 
DataDesc::getType(void) const {
    return mType;
}

#endif  //  MONSERVER_HH
