/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef HTML_ATTRIB_HH
#define HTML_ATTRIB_HH

#include <time.h>
#include <memory>
#include <map>
#include <string>

namespace html {
    class writer;

    /**  Abstract attribute API.
      *  @memo Attribute API.
      *  @author J. Zweizig
      *  @version 1.1; Modified May 4, 2001.
      *  @ingroup IO_html
      */
    class attrib {
    public:
        /**  Null attribute constructor.
	  *  @memo Default constructor.
	  */
        attrib(void);

        /**  Attribute copy constructor.
	  *  @memo Copy Constructor.
	  *  \param x Attribute instance to be copied.
	  */
        attrib(const attrib& x);

        /**  Attribute destructor.
	  *  @memo Destructor.
	  */
        virtual ~attrib(void);

        /**  Make an exact copy of the current attribute.
	  *  @memo Clone an attribute.
	  *  @return Pointer to the cloned attribute.
	  */
        virtual attrib* clone(void) const = 0;

        /**  Append an attribute to the current value separated by a single
	  *  separator character.
	  *  \brief   Append an attribut string to this.
	  *  \param x Attribute instance to be appended to this.
	  *  \param sep separator. 
	  */
        virtual void append(const attrib& x, int sep=' ') = 0;

        /**  Set the writer according to the attribute specified.
	  *  @memo set the writer with an attribute.
	  *  @param w HTML writer to be set.
	  */
        virtual void setWrite(writer& w) const;

        /**  Test whether the attribute is set to its default value.
	  *  @memo Test for default attribute.
	  *  @return True if the attribute is the default value.
	  */
        virtual bool isDefault(void) const;

        /**  Test whether the attribute is set to a null value.
	  *  @memo Test for null attribute.
	  *  @return True if the attribute is the null value.
	  */
        virtual bool isNull(void) const {return false;}

        /**  Test whether the attribute takes an argument.
	  *  @memo Test for boolean attribute.
	  *  @return True if the attribute takes an argument.
	  */
        virtual bool hasArg(void) const;

        /**  Write the attribute value to be used inside a tag.
	  *  @memo write the attrubute value to an html writer.
	  *  @param w HTML writer to which the attribute value is written.
	  */
        virtual void putAttr(writer& w) const;

	/**  Return the attribute value as a string.
	  *  \brief Attribute value string.
	  */
	virtual std::string valString(void) const;
    private:
    };

    /**  attr_ptr class is a copyable and moveable unique_ptr to an attr 
      *  instance.
      */
    class attr_ptr {
    public:
#ifndef __CINT__
	/** Reset the pointer to reference thespecified attribute.
	 *  \brief reset the pointer.
	 *  \param x Attribute to be referenced by the pointer.
	 */
	void reset(attrib* x=0) {_ptr.reset(x);}

	/** Get the pointer value. The pointer keeps the referenced instance.
	 *  \brief Get the pointer value.
	 *  \return pointer to the referenced instance. 
	 */
	attrib* get(void) {return _ptr.get();}

	/** Get the pointer value. The pointer keeps the referenced instance.
	 *  \brief Get the pointer value.
	 *  \return pointer to the referenced instance. 
	 */
	const attrib* get(void) const {return _ptr.get();}

	/** Move constructor for an attribute pointer.
	 *  \brief Move constructor.
	 *  \param x Attribute pointer to be moved.
	 */
	attr_ptr(attr_ptr&& x) = default;

	/** Assign athis pointer to a clone of the instance pointed to by 
	 *  the argument pointer.
	 *  \brief set pointer to a cloned instance.
	 *  \param x   pointer to instance to be cloned.
	 *  \return Referenced to modified instance. 
	 */
	attr_ptr& operator=(const attr_ptr& x);
#else
	void reset(attrib* x) {_ptr = x;}
	attrib* get(void) {return _ptr;}
	const attrib* get(void) const {return _ptr;}
#endif
	/** Construct attribute pointer to cloned attribute
	 *  \brief Clone constructor.
	 *  \param x instance to be cloned
	 */
	attr_ptr(const attrib& x) {reset(x.clone());}

	/** construct an attribute pointer by adopting an existing instance.
	 *  \brief adoption constructor.
	 *  \param ptr pointer to the instance to be adopted.
	 */
	attr_ptr(attrib* ptr = 0) {reset(ptr);}

	/** Construct attribute pointer to cloned dereferenced attribute.
	 *  \brief Clone constructor.
	 *  \param x pointer to instance to be cloned
	 */
	attr_ptr(const attr_ptr& x) {reset(!x ? 0 : x->clone());}

	/** Dereference a constant attribute pointer.
	 *  \brief Dereference pointer.
	 *  \return Pointer to attribute.
	 */
	const attrib* operator->(void) const {return get();}

	/** Dereference an attribute pointer.
	 *  \brief Dereference pointer.
	 *  \return Pointer to attribute.
	 */	
	attrib* operator->(void) {return get();}

	/** Dereference a constant attribute pointer.
	 *  \brief Dereference pointer.
	 *  \return Reference to attribute.
	 */
	attrib& operator*(void) {return *_ptr;}

	/** Dereference a constant attribute pointer.
	 *  \brief Dereference pointer.
	 *  \return Reference to attribute.
	 */
	const attrib& operator*(void) const {return *_ptr;}

	/** Cast pointer to boolean. The return value is true if the pointer 
	 *  is non null.
	 *  \brief Boolean cast operator.
	 *  \return true if the pointer is non-null
	 */
	operator bool(void) const {return bool(_ptr);}
    private:
#ifndef __CINT__
	std::unique_ptr<attrib> _ptr;
#else
	attrib* _ptr;
#endif
    };

    /**  An attribute list is a map associating a set of attribute names 
      *  to with their appropriate values.
      *  @memo Attribute list class.
      *  @author J. Zweizig
      *  @version 1.0; Modified January 5, 2001
      *  @ingroup IO_html
      */ 
    class attList {
    public:

        /**  Construct an empty attribute list.
	  *  @memo Default constructor.
	  */
        attList(void);

        /**  Copy an attribute list by cloning all known attributes.
	  *  @memo Copy constructor.
	  *  @param x Attribute list to be copied.
	  */
        attList(const attList& x) = default;

        /**  Delete all attribute values in the attribute list and clear 
	  *  the list.
	  *  @memo destructor.
	  */
        virtual ~attList(void);

        /**  The specified attribute value is cloned and added to the 
	  *  attribute list with the given name. 
	  *  @memo Add an attribute to the list.
	  *  @param name Attribute name to be added.
	  *  @param a    Value of attribute to be added
	  *  @return reference to the attribute value entry.
	  */
        virtual attrib& addAttr(const std::string& name, const attrib& a);

        /**  The specified attribute value is appended to an existing 
	  *  attribute with the given name. The attribute strings are 
	  *  separated by the separator character. If the named attribute is
	  *  not already in the list or is empty, the argument attribute is 
	  *  added with no separator. 
	  *  @memo Add an attribute to the list.
	  *  @param name Attribute name to be appended.
	  *  @param a    Value of attribute to be appended
	  *  \param sep  Separaion character. If sep is \<null\>, no separator
	  *              is inserted.
	  *  @return reference to the attribute value entry.
	  */
        attrib& appendAttr(const std::string& name, const attrib& a,
			   int sep=' ');

        /**  The specified attribute value is used to replace the 
	  *  attribute with the given name. 
	  *  @memo Replace an attribute in the list.
	  *  @param name Attribute name to be replaced.
	  *  @param a    Replacement attribute value.
	  *  @return reference to the attribute value entry.
	  */
        attrib& repAttr(const std::string& name, const attrib& a);

        /**  Remove the specified attribute from the list.
	  *  @memo remove an attribute.
	  *  @param name Name of attribute to be removed.
	  */
        void remAttr(const std::string& name);

        /**  Delete and destroy all attributes from the list.
	  *  @memo Clear attribute list.
	  */
        void clear(void);

        /**  Test if the list has any attributes.
	  *  @memo test for empty list.
	  *  @return true if the list is empty.
	  */
        bool empty(void) const;

        /**  Test whether the named attribute is included in the list.
	  *  @memo Test if attribute exists.
	  *  @return true if the named attribute is in the list.
	  *  @param key Name of attribute to search for.
	  */
        bool exists(const std::string& key) const;

        /**  Look for the specified attribute. Return a reference if it is 
	  *  found.
	  *  @memo Find a named attribute.
	  *  @return Reference to the found attribute instance.
	  *  @param key name of the attribute to be searched for.
	  */
        const attrib& find(const std::string& key) const;

        /**  Test whether all specified attributes have their default values.
	  *  @memo Test for default list.
	  *  \return true if ALL attributes have their default values.
	  */
        bool isDefault(void) const;

        /**  List all non-default attributes to the HTML document stream.
	  *  @memo List attributes.
	  *  @param w HTML document stream to receive attribute list.
	  */
        virtual void putAttr(writer& w) const;

        /**  Set HTML document writer to reflect any state attributes.
	  *  @memo Set writer attributes.
	  *  @param w Writer to modify based on attribute values.
	  */
        virtual void setWrite(writer& w) const;

        /**  Merge specified attribute list into the current list. Any
	  *  attribute in the specified list that is not covered by the 
	  *  current list will be added to the current list.
	  *  @memo Merge attribute lists.
	  *  @param x List to be merged into the current list.
	  */
        virtual void merge(const attList& x);

    private:
        typedef std::map<std::string, attr_ptr> AttrList;
        typedef AttrList::iterator attr_iter;

    public:
        /**  Constant iterator for use in traversing the attribute list.
	  *  @memo Constant attribute iterator.
	  */
        typedef AttrList::const_iterator const_attr_iter;

        /**  Fetch an iterator pointing to the start of the attribute list.
	  *  @memo List start iterator.
	  *  \return Iterator pointing to the first sequential attribute.
	  */
        const_attr_iter begin(void) const;

        /**  Fetch an iterator pointing to the end of the attribute list.
	  *  @memo List end iterator.
	  *  \return Iterator pointing to the end of the attribute list..
	  */
        const_attr_iter end(void) const;

    private:
        AttrList mList;
    };

}

inline html::attr_ptr&
html::attr_ptr::operator=(const attr_ptr& x) {
    if (x) reset(x->clone());
    else   reset(0);
    return *this;
}

inline bool
html::attList::empty(void) const {
    return mList.empty();
}

inline html::attList::const_attr_iter 
html::attList::begin(void) const {
    return mList.begin();
}

inline html::attList::const_attr_iter 
html::attList::end(void) const {
    return mList.end();
}

#endif  // HTML_ATTRIB_HH
