/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_COMPOUND_HH
#define HTML_COMPOUND_HH

#include "html/Attrib.hh"
#include "html/object.hh"
#include <list>

namespace html {

   /**  compound is an abstract base class for compound html objects. Compound 
    *  objects are ordered HTML object containers that can hold one of more 
    *  simple or compound sub-objects. In addition to a list of constituent
    *  objects, the compound object contains an attribute list that is
    *  valid through the scope of the compound. 
    *  \brief Compound object container.
    *  \author John Zweizig
    *  \version 1.2; Modified April 10, 2019
    *  \ingroup IO_html
    */
   class compound : public object {
   public:
      /**  Construct an empty  compound.
       *  \brief Null constructor.
       */
      compound(void);

      /**  Construct a compound that is identical to the argument. All
       *  constituents of the argument are cloned and added into the 
       *  new compound.
       *  \brief Copy constructor.
       *  \param x compound to be copied.
       */
      compound(const compound& x);

      /**  Destroy the compound object and all its constituent objects.
       *  \brief Compound destructor.
       */
      virtual ~compound(void);

      /**  Set the compound scope attributes and Write out all the consituent 
       *  objects.
       *  \brief Write the compound.
       *  \param out HTML writer to receive the compound and its constituents.
       */
      virtual void write(writer& out) const;

      /**  A new compound object is created from clones of all attributes and 
       *  constituent objects found in the original.
       *  \brief clone the compound object
       *  \return pointer to the clone.
       */ 
      virtual compound* clone(void) const;

      /**  Get the tag name to be used to describe this compound object.
       *  \brief Get the compound tag name.
       *  \return Pointer to a character string containing the tag name.
       */
      virtual const char* getTag(void) const;

      /**  The argument object is cloned and the clone is added to the 
       *  end of the compound object list. A reference to the cloned object
       *  is returned.
       *  \brief Add the specified object to the compound.
       *  \param thing Object to be added to the compound.
       *  \return Reference to the cloned object in the compound.
       */
      object& addObject(const object& thing);

      /**  The argument object is added to the end of the compound object
       *  list. Ownership of the object is taken by the compound. A pointer
       *  to the object is returned.
       *  \brief Add the specified object to the compound.
       *  \param thing Object to be added to the compound.
       *  \return Pointer to the object added to the compound.
       */
      object* addObject(object* thing);

      /** Install any requested script or style files.
       *  \brief Install local files.
       *  \param from local library diretory
       *  \param to   html document directory.
       */
      void install(const char* from, const char* to) const;

      /**  Clone an html object and add the resulting instance to the 
       *  compound list.
       *  \brief add an object to the list.
       *  \param thing Object to be cloned and added.
       *  \return reference to the resulting compound.
       */
      compound& operator<<(const object& thing);

      /**  Add an object instance to the compound list.
       *  \brief add an object to the list.
       *  \param thing pointer to the instance to be added.
       *  \return Reference to the resulting compound.
       */
      compound& operator<<(object* thing);

      /**  Add a text object containing the specified string.
       *  \brief Add a text string to the list.
       *  \param txt text string reference.
       *  \return Reference to the resulting compound.
       */
      compound& operator<<(const std::string& txt);

      /**  Add an html attribute to the  compound attribute list.
       *  \brief add an attribute to the list.
       *  \param thing Attribute to be cloned and added.
       *  \return reference to the resulting compound.
       */
      compound& operator<<(const attrib& thing);

      /**  Get a reference to the last object in the compound object list.
       *  \brief Get back object reference.
       *  \return Reference to the last object in the list.
       */
      object& back(void);

      /**  Get a constant reference to the last object in the compound 
       *  object list.
       *  \brief Get back object reference.
       *  \return Constant reference to the last object in the list.
       */
      const object& back(void) const;

      /**  Get a reference to the first object in the compound object list.
       *  \brief Get front object reference.
       *  \return Reference to the first object in the list.
       */
      object& front(void);

      /**  Get a constant reference to the first object in the compound 
       *  object list.
       *  \brief Get front object reference.
       *  \return Constant reference to the first object in the list.
       */
      const object& front(void) const;

      /**  Add a line-break marker.
       *  \brief Insert a line-break.
       *  \return reference to the inserted line-break object. 
       */
      compound& lineBreak(void);

      /**  Remove all objects from the constituent list.
       *  \brief clear the compound.
       */
      void clear(void);

      /** Test whether html list is empty
       *  \brief Test empty.
       *  \return True if there are no html objets in the list
       */
      bool empty(void) const;

      /**  The specified object us removed from the constitent list and deleted.
       *  \brief remove a constituent object from a compound.
       *  \param thing Pointer to the object to be removed.
       */
      void remObject(const object* thing);

      /**  A clone of the specified attribute is added to the compound scope 
       *  attribute list with the given name.
       *  \brief Add an attribute.
       *  \param name Name of attribute to be added.
       *  \param a    Attribute to be added to the compound scope.
       */
      void addAttr(const std::string& name, const attrib& a);

      /**  A generic attribute with the value string is added to the compound 
       *  scope attribute list with the given name.
       *  \brief Add an attribute.
       *  \param name  Name of attribute to be added.
       *  \param value Attribute to be added to the compound scope.
       */
      void addAttr(const std::string& name, const std::string& value);

      /**  Get a constant reference to the compound attribute list.
       *  \brief Reference attribute list.
       *  \return constant reference to the attribute list.
       */
      const attList& attributes(void) const;

      /**  Get a constant reference to the compound attribute list.
       *  \brief Reference attribute list.
       *  \return constant reference to the attribute list.
       */
      attList& attributes(void);
      
      /**  Remove all attributess from the compound scope.
       *  \brief clear the compound.
       */
      void clearAttr(void);

      /**  Replace the named attribute in the compound scope with a clone of 
       *  the specified attribute.
       *  \brief Replace an attribute.
       *  \param name Name of attribute to be replaced.
       *  \param a    Replacement attribute.
       */
      void repAttr(const std::string& name, const attrib& a);

      /**  Remove the named attribute.
       *  \brief remove the named attribute
       *  \param name Attribute to be removed from the compound scope.
       */
      void remAttr(const std::string& name);

      /** Set the class attribute to this compound.
       *  \brief Set the class.
       *  \param name class name.
       */
      void setClass(const std::string& name);

   private:
      typedef std::list<object_ptr> HtmlList;
      typedef HtmlList::iterator html_iter;
      typedef HtmlList::const_iterator const_html_iter;

   private:
      attList  mAttr;
      HtmlList mList;
   };

   /**  The block class implements named compounds. The block name is 
    *  used as an HTML tag when the block is written to the document
    *  file.
    *  \brief Named compound HTML objects.
    *  \author J.Zweizig
    *  \version 1.1; Modified May 4, 2001
    *  \ingroup IO_html
    */
   class block : public compound {
   public:
      /**  Construct a block. 
       *  \brief Constructor
       *  \param tag Name of the block.
       */
      block(const std::string& tag) : mTag(tag) {}

      /**  Destroy a block.
       *  \brief Destructor.
       */
      virtual ~block(void) {}

      /**  Create an exact duplicate of the current block. This method
       *  pre-empts the HTML object clone virtual method.
       *  \brief Clone a block.
       *  \return Pointer to the newly created copy of the block.
       */
      virtual block* clone(void) const {return new block(*this);}

      /**  Get the tag name of this block.
       *  \brief get the tag.
       *  \return Constant pointer to a static string containing the block name.
       */
      const char* getTag(void) const {return mTag.c_str();}

   private:
      std::string mTag;
   };

   /**  The div class implements the html division (div) block as a 
    *  specialization of compound.
    *  \brief HTML division objects.
    *  \author J.Zweizig
    *  \version 1.1; Modified Apr 4, 2019
    *  \ingroup IO_html
    */
   class div : public compound {
   public:
      /**  Construct a block. 
       *  \brief Constructor
       */
      div(void) {}

      /**  Construct a division with a class name. 
       *  \brief Constructor
       *  \param divclass Class name of this division.
       */
      div(const std::string& divclass);

      /**  Destroy a block.
       *  \brief Destructor.
       */
      virtual ~div(void) {}

      /**  Create an exact duplicate of the current block. This method
       *  pre-empts the HTML object clone virtual method.
       *  \brief Clone a block.
       *  \return Pointer to the newly created copy of the block.
       */
      virtual div* clone(void) const {return new div(*this);}

      /**  Get the div tag string.
       *  \brief get the tag.
       *  \return Constant pointer to a static string containing the block name.
       */
      const char* getTag(void) const {return "div";}
   };

}

//======================================  Inline methods.
inline void 
html::compound::addAttr(const std::string& name, const attrib& a) {
   mAttr.addAttr(name, a);
}

inline html::object*
html::compound::addObject(object* obj) {
   mList.push_back(obj);
   return obj;
}

inline html::object&
html::compound::addObject(const object& obj) {
   return *addObject(obj.clone());
}

inline html::object&
html::compound::back(void) {
   return *mList.back();
}

inline const html::object&
html::compound::back(void) const {
   return *mList.back();
}

inline bool
html::compound::empty(void) const {
   return mList.empty();
}

inline html::object&
html::compound::front(void) {
   return *mList.front();
}

inline const html::object&
html::compound::front(void) const {
   return *mList.front();
}

inline void 
html::compound::repAttr(const std::string& name, const attrib& a) {
   mAttr.repAttr(name, a);
}

inline void 
html::compound::remAttr(const std::string& name) {
   mAttr.remAttr(name);
}

inline html::compound& 
html::compound::operator<<(const object& thing) {
   addObject(thing);
   return *this;
}

inline html::compound& 
html::compound::operator<<(object* thing) {
   addObject(thing);
   return *this;
}

inline const html::attList&
html::compound::attributes(void) const {
   return mAttr;
}

inline html::attList&
html::compound::attributes(void) {
   return mAttr;
}
#endif  // HTML_COMPOUND_HH
