/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_DOCUMENT_HH
#define HTML_DOCUMENT_HH

#include "html/Attrib.hh"
#include "html/compound.hh"
#include "html/head.hh"
#include "html/style.hh"
#include <string>

namespace html {
   class color;
   class writer;

   /**  The HTML document class defines the top level of an HTML document. 
     *  A document is a compound HTML object, and includes in addition to
     *  the usual list of contained objects and attributes, a style directory
     *  and a few meta-data values (\a e.g. the title, refresh rate, etc.).
     *  @memo HTML document class.
     *  @author J. Zweizig
     *  @version 1.0; Modified January 11, 2001
     *  @ingroup IO_html
     */
   class document : public compound {
   public:
      /**  Construct an empty document.
        *  @memo Default constructor.
        */
      document(void);

      /**  Construct an empty document. Preset the title.
        *  @memo Titled document constructor.
        *  @param title  Document title string.
        */
      document(const char* title);

      /**  Destroy a document.
        *  @memo HTML document destructor.
        */
      ~document(void);

      /**  Copy a document into new storage and return the resulting pointer.
        *  @memo clone a document.
	*  @return Pointer to the clone of this object.
	*/
      document* clone(void) const;

      /**  Get the tag name to be used. In this case it is document.
        *  \brief Get tag character string.
	*  \return Pointer to constant tag character string.
	*/
      const char* getTag(void) const;

      /**  Get a constant reference to the header section.
        *  \brief referecne header section.
	*  \return Constant reference to the header section.
	*/
      const head& header(void) const;

      /**  Get a reference to the header section.
        *  \brief reference header section.
	*  \return reference to the header section.
	*/
      head& header(void);

      /** Initialze the header structure.
       */
      void init_header();

      /**  Install scripts and style (css) files.
        *  \brief Install scripts and styles.
	*  \param from Local system html scripts directory.
	*  \param to   Local directory to receive document.
	*/
      void install(const char* from, const char* to) const;
      
      /**  Write out the document using recursive calls to the specified html
        *  writer.
	*  \brief Write the document.
	*  \param out Reference to the html writer. 
	*/
      void write(writer& out) const;

      /**  Set the document background color.
        *  \brief Set the background color.
	*  \param color Color specifier.
	*/
      void setBackgroundColor(const color& color);

      /**  Set the document refresh time. A refresh time of zero indicates
        *  that the document is not to be refreshed.
	*  @memo specify a refresh period.
	*  \param t Refresh time in seconds.
	*/
      void setRefresh(double t);

      /**  Set the document title string.
        *  \brief Set the document title.
	*  \param title New title string.
	*/
      void setTitle(const std::string& title);

   private:
      std::string mTitle;
      style       mStyle;
      double      mRefresh;
      head        mHead;
   };

   //===================================  inline functions
   inline const head&
   document::header(void) const {
      return mHead;
   }

   inline head&
   document::header(void) {
      return mHead;
   }
}
#endif  // HTML_DOCUMENT_HH
