/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_HEAD_HH
#define HTML_HEAD_HH

#include "html/compound.hh"
#include "html/tag_only.hh"
#include "html/text.hh"
#include <string>

namespace html {

   class writer;

   /** The head class contains the html document header data, including 
    *  a title, meta-data, related data (scripts and styles) links and 
    *  global style defintions.
    */ 
   class head : public compound {
   public:
      /** Construct an empty header section.
       *  \brief defaul constructor.
       */
      head(void);

      /** Destroy the header section.
       *  \brief Destructor.
       */
      virtual ~head(void) = default;

      /** Clone this header.
       *  \brief Clone header.
       *  \return Pointer to cloned header.
       */
      head* clone(void) const;

      /** Define the tag name to be used for the html block ("head").
       *  \brief Head section tag.
       *  \return Tag string.
       */
      const char* getTag(void) const {return "head";}

      /** write out the header data. This uses the compund writer for all the
       *  included data, but makes sure that the head tags are each on new
       *  lines.
       */
      void write(writer& w) const;
   };

   /** Define an html related data link. This is primarily used to define 
    *  the location (URL) of a css file containing global style information.
    *  \brief Related data link
    */
   class rel_link : public tag_only {
   public:
      /** Construct a related data link object. By default the relation and
       *  data type are appropriate for a css style file. If the install() 
       *  method is run for a document, the requested data files will be 
       *  copied from the specified system library to the targei html 
       *  directory. 
       *  \brief Define related data URL.
       *  \param href URL of the related data.
       *  \param rel  Related data sheet to receive linkd data.
       *  \param type Type of data in target file.
       */
      rel_link(const std::string& href, const std::string& rel="stylesheet",
	       const std::string& type="text/css");
      /** Related data link destructor.
       * \brief Destructor.
       */
      virtual ~rel_link(void) = default;

      /** Clone a related link.
       *  \brief clone this object
       *  \return pointer to the cloned object.
       */
      rel_link* clone(void) const;

      /** Copy a file from the system html data directory to the target 
       *  directory for this document.
       *  \brief install a data frile.
       *  \param from Path for system html data.
       *  \param to   Path of document
       */
      void install(const char* from, const char* to) const;
   };

   /** Header metadata descriptor. The header metadata is encoded as a tag with
    *  name and contents attributes.
    *  \brief Header meta-data
    */
   class meta : public tag_only {
   public:
      /** Construct a metadata tag with the specified name and contents.
       *  \brief Header metadata constructor.
       *  \param name    Metadata name
       *  \param content Metadata content.
       */
      meta(const std::string& name, const std::string& content);

      /** metadata destructor.
       *  \brief Destructor.
       */
      virtual ~meta(void) = default;

      /** Clone the metadata descriptor.
       *  \brief Clone metadata.
       *  \return Pointer to the cloned metadata.
       */
      meta* clone(void) const;
   };

   /** Document header title information.
    *  \brief Document title.
    */
   class title : public text_block {
   public:
      /** Construct the document title text.
       *  \brief Title constructor.
       *  \param docttl Ducument title string.
       */
      title(const std::string& docttl);
      
      /** Title destructor.
       *  \brief Destructor.
       */
      virtual ~title(void) = default;
   };
}

#endif // !defined(HTML_HEAD_HH)
