/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_OBJECT_HH
#define HTML_OBJECT_HH
#include <vector>
#include <memory>

/**  The %html namespace  contains the basic classes needed to construct an
  *  html document. all html elements are based on the abstract html::object 
  *  class that includes methods for writing and cloning the html element.
  *  Complex elements are based on the abstract html::compound class that 
  *  contains a tag name, a list of tag attributes and a list of elements 
  *  that are contained within the compound.
  *
  *  A document is constructed as follows:
  *  \verbatim
      html::document doc("Coherence Monitor Results");
      doc.header() << script("text/javascript", "sorttable.js");
      doc.header() << script("text/javascript", "collapsible.js");
      doc.lineBreak();
      doc << "Text data may be entered into a compound like a document "
          << "directly with the shift operator. Other object can be added "
	  << "to a compound with either the compound::addObject() method "
	  << "or the equivalent shift operator. if the argument of addObject "
	  << "is a non-const object pointer, ownership is passed to the "
	  << "compound, e.g. todays date is: " << new html::text(Now(), fmt);  
      html::writer out("my_file.html");
      doc.write(out);
     \endverbatim
  *  \brief  HTML document definition classes.
  *  \author John Zweizig (john.zweizig(at)ligo.org)
  */
namespace html {
  class writer;
  typedef int dim_type;

  /**  The HTML object class is an abstract class defining the API for any 
    *  HTML object to be displayed.
    *  @memo Abstract HTML Object API.
    *  @author   John Zweizig
    *  @version  1.0; Modified January 5, 2001
    *  @ingroup IO_html
    */
  class object {
  public:
    /**  Null object constructor
      */
    object(void);

    /**  Object destructor
     */
    virtual ~object(void);

    /**  Create an identical copy of the object. The clone is owned by
      *  the calling function and will not be deleted when the original
      *  is destroyed.
      *  @memo Clone an object.
      *  @return Pointer to the object clone.
      */
    virtual object* clone(void) const = 0;

    /**  Get the horizontal size of a printable object.
      *  @memo Get the object size in x.
      *  @return Horizontal size in points.
      */
    virtual dim_type getXSize(void) const;

    /**  Get the vertical size of a printable object.
      *  @memo Get the object size in y.
      *  @return Vertical size in points.
      */
    virtual dim_type getYSize(void) const;

    /**  Install scripts and css files from the default system location to 
      *  the document target directory.
      *  \brief Copy scripts to default directory.
      *  \param from Default scripts directory available to document writer. 
      *  \param to   Scripts directory used by html document. 
      */
    virtual void install(const char* from, const char* to=0) const;

    /**  Write the description of this object and all included objects
      *  in the case of a compound to an html file.
      *  @memo Write out the object.
      *  @param out HTML write to receive the object description.
      */
    virtual void write(writer& out) const;
  };

    /** The object_ptr class points to an html object. It is copyable via the 
     *  object clone method and moveable.
     */
    class object_ptr : public std::unique_ptr<object> {
    public:
	/** Construct an empty html object pointer
	 *  \brief Default constructor.
	 */
	object_ptr(void) {}

	/** Construct an html object pointer pointing to the specified object
	 *  \brief Construct object pointer to an existing object.
	 *  \param obj Pointer to object to be adopted.
	 */
	object_ptr(object* obj) {
	    reset(obj);
	}

	/** Construct an html object pointer pointing to the specified object
	 *  \brief Construct object pointer to an existing object.
	 *  \param obj Reference to object to be cloned.
	 */
	object_ptr(const object& obj) {
	    reset(obj.clone());
	}
	
	/** Construct an html object pointer the points to a clone of the 
	 *  specified object.
	 *  \brief Construct from a clone of a dereferenced pointer.
	 *  \param p Pointer to object to be cloned.
	 */
	object_ptr(const object_ptr& p) {
	    reset(p ? p->clone() : 0);
	}
	
	/** Move contents of an html object pointer to a new pointer.
	 *  \brief Move constructor.
	 *  \param o Pointer to be moved.
	 */
	object_ptr(object_ptr&& o) = default;

	/** Delete the referenced html object and destroy the object pointer
	 *  \brief Destructor
	 */
	virtual ~object_ptr(void) {}

	/** Copy an object pointer to en existing pointer by cloning the 
	 *  referenced object.
	 *  \brief Copy pointer by cloning the object. 
	 *  \param p Pointer to be copied.
	 *  \return Reference to modified pointer.
	 */
	object_ptr& operator=(const object_ptr& p) {
	    reset(p->clone());
	    return *this;
	}

	/** Move an object pointer to an existing pointer.
	 *  \brief Move pointer to an existing pointer instance.
	 *  \param p Pointer to be moved.
	 *  \return Reference to modified pointer.
	 */
	object_ptr& operator=(object_ptr&& p) = default;

	/** Adopt an object instance to an existing pointer.
	 *  \brief Adopt a specied instance.
	 *  \param p Pointer to the instance to be adopted.
	 *  \return Reference to modified pointer.
	 */
	object_ptr& operator=(object* p) {
	    reset(p);
	    return *this;
	}
     };

    typedef std::vector<object_ptr> html_vect;
    typedef html_vect::iterator html_vect_iter;
    typedef html_vect::const_iterator const_html_vect_iter;
}

#endif  // HTML_OBJECT_HH
