/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_SCRIPT_HH
#define HTML_SCRIPT_HH

#include "html/Attrib.hh"
#include "html/compound.hh"
#include <string>

namespace html {

   /**  The script class defines an atomic html object containing a script
     *  string and a list of script-specific attributes. A script source file
     *  may be specified in lieu of the script text.
     *  @memo Html script object
     *  @author J. Zweizig
     *  @version 1.1; Modified April 4, 2019
     *  @ingroup IO_html
     */
   class script : public compound {
   public:
      /**  Construct an empty script object.
        *  @memo Default constructor.
	*/
      script(void);

      /**  Construct a script object and set its type and, optionally, the
        *  source URL.
	*  \brief Data constructor
	*  \param type   String containing the script langiage.
	*  \param source Script source URL.
	*/
      script(const std::string& type, const std::string& source="");

      /**  Destroy a script object.
        *  @memo Script destructor.
	*/
      virtual ~script(void);

      /**  Create a script object identical to the existing instance and return
        *  a pointer to the result. This method overrides the default action 
	*  of the html object clone virtual method.
	*  @memo Clone a script object.
	*  @return pointer to an exact copy of the current script instance.
	*/
      script* clone(void) const;

      /**  Specify the tag string.
        *  \brief Tag string.
	*  \return Tag character string.
	*/
      const char* getTag(void) const;

      /**  Copy the object script to a location available to the web browser.
        *  @memo Install the script.
	*  @param from local system script library location.
	*  @param to   Web accessible document directory.
	*/
      void install(const char* from, const char* to) const;

      /**  Replace any existing script with the specified string.
        *  @memo Set the script.
	*  @param txt String containg new script.
	*/ 
      void insertScript(const std::string& txt);

      /**  Append a specified string to any existing script on the object.
        *  @memo Append a script string.
	*  @param txt String to be added to the script.
	*/
      void addScript(const std::string& txt);

      /**  Append a specified string to any existing script in the object.
        *  @memo Append a string to the script.
	*  @param txt String to be added to the script.
	*  @return Reference to the updated script object.
	*/
      script& operator<<(const std::string& txt);

      /**  Set the location of the script source file.
        *  \brief Set the source location.
	*  \param path Location of the script.
	*  \return Reference to the script instance.
	*/
      void setSource(const std::string& path);

      /**  Specify the script type attribute.
        *  @memo Set the script type.
	*  @return Script type string.
	*  @param lang A script type (language) specifier.
	*/
      void setType(const std::string& lang);
   };

   //===================================  Inline methods.
   inline const char*
   script::getTag(void) const {
      return "script";
   }
}

#endif  // HTML_SCRIPT_HH
