/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef HTML_STYLE_HH
#define HTML_STYLE_HH

#include <string>
#include "html/compound.hh"

namespace html {
    class color;
    class font;
    class size;
    class writer;
    class textAttr;

    /** The style attrribute is a series of name:value pairs that specify
     *  the values for various attributes to be used in the appropriate
     *  context. The style is specified either within a style attribute for 
     *  a specific html element, or within a style script from a css file.
     */
    class styleAttrib : public attrib {
    public:
	/** Construct a style attribute string.
	 *  \brief style attribute constructor.
	 */
	styleAttrib(void);

	/** Construct a style attribute string.
	 *  \brief Style attribute constructor.
	 *  \param attrs Style attribute string.
	 */
	styleAttrib(const std::string& attrs);

	/** Construct a style attribute from a text attribute list.
	 *  \brief style attribute constructor.
	 *  \param attrs Text attribute list.
	 */
	styleAttrib(const textAttr& attrs);

	/** Destroy a style attribute. object.
	 *  \brief Destructor.
	 */
	~styleAttrib(void) = default;

	/** Clone a style attribute.
	 *  \brief clone style attribute.
	 *  \return pointer to the cloned instance.
	 */
	styleAttrib* clone(void) const;

	/**  Add an attribute and its value to the document.
	  *  \brief Add an attribute.
	  *  \param name Attribute name
	  *  \param attr Attribute value.
	  */
	void addAttr(const std::string& name, const attrib& attr);

	/**  Add an attribute and its value to the document.
	  *  \brief Add an attribute.
	  *  \param attr Attribute to be appended.
	  *  \param sep  separator character.
	  */
	void append(const attrib& attr, int sep=' ');

	/**  Set the style color attribute.
	  *  @brief Set the color attribute.
	  *  @param c Color attribute.
	  *  @return Reference to this StyleClass object.
	  */
	styleAttrib& setColor(const color& c);

	/**  Set the style font face attribute.
	  *  @brief Set the font attribute.
	  *  @param c Font attribute.
	  *  @return Rference to this StyleClass object.
	  */
	styleAttrib& setFace(const font& c);

	/**  Set the style size attribute.
	  *  @brief Set the size attribute.
	  *  @param c Size attribute.
	  *  @return Reference to this StyleClass object.
	  */
	styleAttrib& setSize(const size& c);

	/**  Set the style weight attribute.
	  *  @brief Set the weight attribute.
	  *  @param c Weight attribute.
	  *  @return Reference to this StyleClass object.
	  */
	styleAttrib& setWeight(const std::string& c);

	/**  Get the combined style attribute string
	  *  \brief Style value string.
	  *  \return Style string.
	  */
	std::string valString(void) const;
    private:
	std::string mStyle;
    };	

    /**  Class style contains text style attributes.
      *  @memo Text style attribute.
      *  @author John Zweizig
      *  @version 1.1; Last modified March 3, 2008
      *  @ingroup IO_html
      */
    class StyleClass : public object {
    public:
	/**  Construct a class style specifier.
	  *  \brief Constructor
	  *  \param name Name of html class to which this style is applied.  
	  */
	StyleClass(const std::string& name, const styleAttrib& attrib);

	/**  Destroy a style specifier.
	  *  @brief Destructor.
	  */
	~StyleClass(void) = default;

	/** Clone a class style object.
	 *  \brief clone this object.
	 *  \return pointer to cloned class style object
	 */
	StyleClass* clone(void) const;

	/**  Write an attribute to the a document.
	  *  @brief Write an attribute.
	  *  @param w Writer to whic the attribute will be written.
	  */
	void write(writer& w) const;
    private:
	std::string class_name;
	styleAttrib style_attr;
    };

    /**  Class style contains a dictionary of text style attributes.
      *  @memo Text style attribute.
      *  @author J. Zweizig
      *  @version 1.1; Last modified March 3, 2008
      *  @ingroup IO_html
      */
    class style : public compound {
    public:
	/**  Construct an empty style dictionary.
	  *  @brief Defult constructor
	  */
	style(void);

	/**  Destroy a style dictionary.
	  *  @brief Destructor.
	  */
	~style(void) = default;

	/** Clone a style object.
	 *  \brief Clone a style.
	 *  \return Pointer to cloned object.
	 */
	style* clone(void) const;

	/**  Add a style specifier for the given class name to the dictionary.
	  *  @brief Add a style class.
	  *  @param name Style class name.
	  *  @param c    Named style class.
	  */
	void addClass(const StyleClass& c);

	/** Get the style tag string.
	 *  \brief style tag string.
	 *  \return pointer to the string "style".
	 */
	const char* getTag(void) const;
    };

    //==================================  inline methods.
    inline const char*
    style::getTag(void) const {
	return "style";
    }
}

#endif //  HTML_STYLE_HH
