/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef HTML_TAG_ONLY_HH
#define HTML_TAG_ONLY_HH
#include "html/object.hh"
#include "html/Attrib.hh"

namespace html {
   class writer;
   
   /** class to implement non-data tags that contain only a the tag name and
    *  a series of attribute terms. These terms aer terminated with "/>" and 
    *  require no closing tag or intervening data. 
    *  \brief Non-data html tags.
    */
   class tag_only : public object {
   public:
      /** Construct an instance with the specified tag name.
       *  \brief Constructor.
       *  \param name Tag name.
       */
      tag_only(const std::string& name);

      /** Destroy the tag instance.
       *  \brief Destructor.
       */
      virtual ~tag_only(void);

      /** Construct a new clone of this instance and return a pointer.
       *  \brief Clone this instance.
       *  \return pointer to the new clone instance.
       */
      virtual tag_only* clone(void) const;

      /** Write the instance to a document stream.
       *  \brief Write the instance to a document stream.
       *  \param out Html writer instance.
       */
      virtual void write(writer& out) const;

      /** Add an attribute to the attribute list.
       *  \brief Add an attribute to the list.
       *  \param attr  Attribute name.
       *  \param value Typed attribute value.
       *  \return Reference to added attribute
       */
      attrib& addAttr(const std::string& attr, const attrib& value);

      /** Add a generic attribute woth a specified string value to the 
       *  attribute list.
       *  \brief Add an attribute to the list.
       *  \param attr  Attribute name.
       *  \param value String attribute value.
       *  \return Reference to added attribute
       */
      attrib& addAttr(const std::string& attr, const std::string& value);

      /** Get a constant reference to the tag attributre list.
       *  \brief Attribute list reference.
       *  \return Reference to the attribute list.
       */
      const attList& attributes(void) const;

      /** Get a reference to the tag attributre list.
       *  \brief Attribute list reference.
       *  \return Reference to the attribute list.
       */
      attList& attributes(void);
   private:
      std::string _tag;
      attList     _attrs;
   };

   //===================================  Inline methods.
   inline const attList&
   tag_only::attributes(void) const {
      return _attrs;
   }
   
   inline attList&
   tag_only::attributes(void) {
      return _attrs;
   }

   inline attrib&
   tag_only::addAttr(const std::string& name, const attrib& value) {
	 return _attrs.addAttr(name, value);
   }
}

#endif // !defined(HTML_TAG_ONLY_HH)
