/* -*- mode: c++; c-basic-offset: 3; -*- */

#ifndef JSON_STACK_HH
#define JSON_STACK_HH

#include <json/value.h>
#include <stdexcept>
#include <string>
#include <vector>
#include <iosfwd>

/**  jsonStack is a stack-based search tool for elements in a json document.
  *  jsoncpp provides the internal document parsing and storage while the 
  *  jsonStack class allows the user to select or iterate over certain elements
  *  in a document for conversion and/or unpacking.
  *  \author John Zweizig (john.zweizig@ligo.org)
  *  \version $Id$
  */
class jsonStack {
public:
   /// Boolean vector data typw
   typedef std::vector<bool>        bool_vect;
   /// Numeric vector data typw
   typedef std::vector<double>      dble_vect;
   /// String vector data typw
   typedef std::vector<std::string> string_vect;

public:
   /**  Class to automatically clean a stak by popping a json element off 
     *  the stack when the instance falls out of scope.
     */
   class autopop {
   public:
      /**  Note the stack from which the top element is to be popped when this
        *  instance falls out of scope.
	*  \brief Constructor.
	*  \param js Stack instance to be cleaned.
	*/
      autopop(jsonStack& js) : _js(js) {}

      /**  Pop off the top element when this instance goes away.
       */
      ~autopop(void) {_js.pop();}
   private:
      jsonStack& _js;
   };
   
   
public:
   /**  Construct an empty document stack.
     *  \brief stack constructor.
     */
   jsonStack(void);
   
   /**  Construct a json document stack. Push the document in the specified
     *  file onto the stack. if an error is encountered while reading the file 
     *  or parsing the document, a description of the error is written to 
     *  stderr and a std::runtime_error exception is thrown.
     *  \brief stack constructor.
     *  \param file File containing a json document
     */
   jsonStack(const std::string& file);

   /** Clear the stack.
    */
   void clear(void);
   
   /**  Number of elements pushed onto the current stack.
     *  \brief Stack depth.
     *  \return Numer of elements pushed onto the stack
     */
   size_t depth(void) const;

   /**  Test for empty stack.
     *  \brief Test for empty stack.
     *  \return True if stack is empty.
     */
   bool empty(void) const;

   /**  Fetch a data element and convert it to boolean type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a boolean value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, bool& value) const;

   /**  Fetch a data element and convert it to integer type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as an integer value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, int& value) const;

   /**  Fetch a data element and convert it to double type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a double value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, double& value) const;

   /**  Fetch a data element and convert it to string type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a string value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, std::string& value) const;

   /**  Fetch a data element and convert it to boolean vector type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a boolean vector value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, bool_vect& value) const;

   /**  Fetch a data element and convert it to double vector type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a double vector value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, dble_vect& value) const;

   /**  Fetch a data element and convert it to string vector type.
     *  \brief Fetch a data element.
     *  \param name  Name of the requested element.
     *  \param value Element returned as a string vector value.
     *  \return Number of elements saved.
     */
   size_t fetch_data(const std::string& name, string_vect& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  boolean variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value boolean variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, bool& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as an 
     *  integer variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value boolean variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, int& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  numeric (double) variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value Double variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, double& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  string variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value string variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, std::string& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  boolean vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value boolean vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, bool_vect& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  numeric (double) vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value double vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, dble_vect& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  strinig vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value string vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(const Json::Value& name, string_vect& value) const;

   /**  Convert the top value if necessary and save it as a boolean variable.
     *  \brief Convert and save top value.
     *  \param value boolean variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(bool& value) const;

   /**  Convert the top value if necessary and save it as an 
     *  integer variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param value boolean variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(int& value) const;

   /**  Convert the top value if necessary and save it as a 
     *  numeric (double) variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param value Double variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(double& value) const;

   /**  Convert the top value if necessary and save it as a 
     *  string variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param value string variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(std::string& value) const;

   /**  Convert the top value if necessary and save it as a 
     *  boolean vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param value boolean vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(bool_vect& value) const;

   /**  Convert the top value if necessary and save it as a 
     *  numeric (double) vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value double vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(dble_vect& value) const;

   /**  Convert the specified jsoncpp value if necessary and save it as a 
     *  strinig vector variable.
     *  \brief Convert and save a jsoncpp value.
     *  \param name value to be converted/saved.
     *  \param value string vector variable into which the value will be saved.
     *  \return Number of elements saved.
     */
   size_t fetch_value(string_vect& value) const;
   
   /**  Iterate over the element in the array at the top of the stack. The 
     *  iterate method will fail if the top element is not an array. After 
     *  each time iterate is called, the first/next element is pushed onto the 
     *  top of the stack. If no more elements are available, iterate returns 
     *  a value of false. It is up to the user to pop the element off the 
     *  stack after each successful call to iterate(). This may be done by 
     *  constructing an autopop instance (shown below) in a scope defined 
     *  conditionally by the sucess of iterate() or by adding a pop call with 
     *  each successful iteration.
     *
     *  A typical use case for iterate would be as follows:
     *  \verbatim
            jsonStack jss(file);
	    ...
	    if (jss.push_member("array_name")) {
	       jsonStack::autopop pop_array(jss);
	       while (jss.iterate()) {
	          jsonStack::autopop pop_el(jss);
	          double vbl;
	          jss.fetch_value(vbl);
		  cout << "array value: " << vbl << endl;
               }
            }
        \endverbatim
     *  \brief Iterate over elements in the array on the top of the stack.
     *  \return True if a new element has been pushed onto the top of the stack.
     */
   bool iterate(void);

   /**  Test whether top element is an array.
     *  \brief Test whether top of stack is an array.
     *  \return True if top element is an array.
     */
   bool isArray(void) const;

   /**  Test whether top element is a boolean value.
     *  \brief Test whether top of stack is boolean.
     *  \return True if top element is boolean.
     */
   bool isBool(void) const;

   /**  Test whether the structure on top of the stack contains an member 
     *  with the specified name.
     *  \brief Test for named element.
     *  \param name Name of element to test for.
     *  \return True if named element exists.
     */
   bool isMember(const std::string& name) const;

   /**  Test whether top element is a numeric value.
     *  \brief Test whether top of stack is numeric.
     *  \return True if top element is numeric.
     */
   bool isNumeric(void) const;

   /**  Test whether top element is a structure.
     *  \brief Test whether top of stack is an structure.
     *  \return True if top element is an structure.
     */
   bool isStruct(void) const;

   /**  Get the length of the top element (array or structure).
     *  \brief Length of array or structure
     *  \return Number of elements in the top array.
     */
   size_t length(void) const;

   /**  Get a list of members in the structure at the top of the stack.
     *  \brief List structure members.
     *  \return List of member names.
     */
   string_vect getNames(void) const;

   /**  Read in the json document contained in the specified file. The document
     *  is parsed and pushed onto the top element of the stack. If an error is 
     *  found in reading the file or in the json syntax while parsing the 
     *  document, a description of the error is printed to stderr.
     *  \brief parse A json document from a file.
     *  \param file File path
     *  \return True if a non-empty document was read and parsed.
     */
   bool parse(const std::string& file);

   /**  Read in the json document contained in the specified istream. The 
     *  document is parsed and pushed onto the top element of the stack. If 
     *  an error is found in reading the file or in the json syntax while 
     *  parsing the document, a description of the error is printed to stderr.
     *  \brief parse A json document from a file.
     *  \param stream input stream
     *  \param file name of input file or stream.
     *  \return True if a non-empty document was read and parsed.
     */
   bool parse(std::istream& stream, const std::string& file="");
   
   /**  Push an empty element onto the stack.
     *  \brief Push an empty element.
     *  \return Always true (successful)
     */
   bool push_element(void);
   
   /**  Push the indexed element of an array
     *  \brief Push an array element.
     *  \param index Index of element to be pushed.
     *  \return True if the top element is a structure of length > index.
     */
   bool push_element(size_t index);
   
   /**  Push a specified json value.
     *  \brief Push a specified element.
     *  \param val constant reference to value to be pushed.
     *  \return Always true (successful)
     */
   bool push_element(const Json::Value& val);

   /**  Push the contents of a named structure member onto the stack.
     *  \brief Push a named structrure member.
     *  \param name Name of member to be pushed.
     *  \return true if the named element was found and pushed.
     */
   bool push_member(const std::string& name);

   /**  Pop the top json element off of the stack. This must be called 
     *  explicitly or imloicitly via autopop after each successful  push or
     *  iterate operation.
     *  \brief Remove the top element of the stack.
     *  \return true if the stack is not empty after the pop operation.
     */
   bool pop(void);

   /**  Reset the iterator counter. This may be necessary if an iteration 
     *  loop exited before hitting the last element.
     */
   void reset_iterator(void);

   /**  Check that all specified member names are valid for top structure.
     *  \brief check member names
     *  \param names valid member list
     *  \return string with comma separates list of invalid members.
     */
   std::string validate(const char* names[]) const;

   /**  Return a reference to the top element on the stack.
     *  \brief Top element reference
     *  \return Reference to top element.
     */
   const Json::Value& value(void) const;

private:
   Json::Value& value(void);

private:
   enum eltype {
      k_element,
      k_array,
      k_structure
   };

  struct stack_element {
    stack_element(void);
    stack_element(const Json::Value& se);
    Json::Value _value;
    eltype      _type;
    Json::ArrayIndex _size;
    Json::ArrayIndex _iter;
  };
  std::vector<stack_element> _stack;
};

//======================================  Inline functions
inline size_t
jsonStack::depth(void) const {
  return _stack.size();
}

inline bool
jsonStack::empty(void) const {
  return _stack.empty();
}

inline bool 
jsonStack::isArray(void) const {
   return value().isArray();  
}

inline bool 
jsonStack::isBool(void) const {
   return value().isBool();  
}

inline bool 
jsonStack::isMember(const std::string& name) const {
   return value().isMember(name); 
}

inline bool 
jsonStack::isNumeric(void) const {
   return value().isNumeric();  
}

inline bool 
jsonStack::isStruct(void) const {
   return value().isObject();
}

inline size_t
jsonStack::length(void) const {
  return value().size();
}

inline const Json::Value&
jsonStack::value(void) const {
  if (empty()) throw std::runtime_error("Operation on empty json stack");
  return _stack.back()._value;
}

inline Json::Value& 
jsonStack::value(void) {
  if (empty()) throw std::runtime_error("Operation on empty json stack");
  return _stack.back()._value;
}

inline size_t 
jsonStack::fetch_value(bool& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(int& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(double& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(std::string& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(bool_vect& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(dble_vect& val) const {
   return fetch_value(value(), val);
}

inline size_t 
jsonStack::fetch_value(string_vect& val) const {
   return fetch_value(value(), val);
}


#endif     // !defined(JSON_STACK_HH)
