/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef JSONSTREAM_HH
#define JSONSTREAM_HH

#include "jsonStack.hh"
#include <iosfwd>
#include <memory>

class DVector;
class TSeries;
namespace containers {
  class fSeries;
}

/**  jsonStream is a class intended to stream gds containers to and from json
 *   documents.
 */
class jsonStream {
public:
  ///  Copy double vector data type from jsonStack
  typedef jsonStack::dble_vect   dble_vect;

  ///  Copy string vector data type from jsonStack
  typedef jsonStack::string_vect string_vect;

  ///  Automatic frequency series pointer for leak-proof data passing
  typedef std::unique_ptr<containers::fSeries> fseries_ptr;

  ///  Automatic time series pointer for leak-proof data passing
  typedef std::unique_ptr<TSeries> tseries_ptr;

  ///  Automatic data vector pointer for leak-proof data passing
  typedef std::unique_ptr<DVector> dvector_ptr;
public:
  /** Default constructor.
   */
  jsonStream(void);

  /** Construct a jsonStream instance and fill the stack from a STL istream.
   *  \brief initializing constructor.
   *  \param in STL input stream containing a json document.
   */
  jsonStream(std::istream& in);

  /** Get the length of the array on the top of the stack.
   *  \brief Get array length
   *  \return length of array.
   */
  size_t arrayLength(void) const;

  /** Test for empty stack
   *  \brief Test for empty
   *  \return true if stack is empty.
   */
  bool empty(void) const;

  /** Construct a frequency series from the json structure at the top of 
   *  the stack.
   *  \brief Get a frequency series.
   *  \return Pointer to the resulting series
   */
  fseries_ptr getfSeries(void) const;

  /** Construct a frequency series from the indexed json structure in the 
   *  array at the top of the stack.
   *  \brief Get a frequency series instance.
   *  \param index Index into the array.
   *  \return Pointer to the resulting series instance.
   */
  fseries_ptr getfSeries(size_t index);

  /** Construct a frequency series from the json structure identified  
   *  by the channel name.
   *  \brief Get a frequency series instance.
   *  \param name Name field of the series to be returned.
   *  \return Pointer to the resulting series instance.
   */
  fseries_ptr getfSeries(const std::string& name);

  /** Construct the frequency series with the specified name from an
   *  STL input stream containing a json document.
   *  \brief Read and construct a frequency series instance.
   *  \param in   STL input stream containing the json document.
   *  \param name Name field of the series to be imported.
   *  \return Pointer to the resulting series.
   */
  fseries_ptr getfSeries(std::istream& in, const std::string& name="");

  /** Construct a time series from the json structure at the top of 
   *  the stack.
   *  \brief Get a frequency series instance.
   *  \return Pointer to the resulting series instance.
   */
  tseries_ptr getTSeries(void) const;

  /** Construct a time series from the indexed json structure in the 
   *  array at the top of the stack.
   *  \brief Get a time series instance.
   *  \param index Index into the array of the series to be built.
   *  \return Pointer to the resulting series instance.
   */
  tseries_ptr getTSeries(size_t index);

  /** Construct a time series from the json structure identified  
   *  by the channel name.
   *  \brief Get a time series instance.
   *  \param name Name field of the series to be returned.
   *  \return Pointer to the resulting series instance.
   */
  tseries_ptr getTSeries(const std::string& name);

  /** Construct the time series with the specified name from an
   *  STL input stream containing a json document.
   *  \brief Read and contruct a time series instance.
   *  \param in   STL input stream containing the json document.
   *  \param name Name field of the series to be imported.
   *  \return Pointer to the resulting series instance.
   */
  tseries_ptr getTSeries(std::istream& in, const std::string& name="");

  /** Test whether the top element on the stack is an array.
   *  \brief Test for array.
   *  \return true if the top stack element is a time series structure.
   */
  bool isArray(void) const;

  /** Test whether the structure at the top of the stack is a time series.
   *  \brief Test for time series.
   *  \return true if the top stack element is a time series structure.
   */
  bool isTSeries(void) const;

  /** Test whether the structure at the top of the stack is a frequency series.
   *  \brief Test for frequency series.
   *  \return True if the top element is a frequency series structure.
   */
  bool isfSeries(void) const;

  /** List all the objects in the top array of the json stack by name and type.
   *  \brief List objects at the top of the stack
   *  \param names String vectory to receive the structure names.
   *  \param types String vectory to receive the structure types.
   */
  void listObjects(string_vect& names, string_vect& types) const;

  /** Parse the json document contines in the specified STL input stream and
   *  push it onto the  internal stack.
   *  \brief Parse a json document.
   *  \param in STL input stream containing the json document to be parsed.
   */
  void parse(std::istream& in);

  /** Write the argument frequency series to the specified STL output stream 
   *  as a json structure.
   *  \brief Write a frequency series json structure.
   *  \param fs  Frequency series to be written.
   *  \param out STL output stream to resceive the json element.
   */
  void putfSeries(const containers::fSeries& fs, std::ostream& out);

  /** Write the argument time series to the specified STL output stream 
   *  as a json structure.
   *  \brief Write a time series json structure.
   *  \param ts  Time series to be written.
   *  \param out STL output stream to resceive the json element.
   */
  void putTSeries(const TSeries& ts, std::ostream& out);

private:
  /** Internal function to construct a data vector from the top element of 
   *  the json stack.
   *  \brief Construct a data vector as a json data array element.
   *  \param doc jsoncpp structure to receive vector type and data array.
   *  \param dv  Constant reference to data vector to be streamed.
   */
  dvector_ptr getDVector(void) const;
  
  /** Internal function to write a data vector as a data type string
   *  with key name "data-type", and a double array with key name "data".
   *  Complex data is written as a double array with two elements for each
   *  complex value.
   *  \brief Write a %DVector as elements in a json structure.
   *  \param doc jsoncpp structure to receive vector type and data array.
   *  \param dv  Constant reference to data vector to be streamed.
   */
  void        putDVector(Json::Value& doc, const DVector& dv) const;
  
  /** Push the structure corresponding to the specified type and name onto 
   *  the top of the json stack. Select returns true if the requested structure
   *  was pushed on to the top of the stack and throws an exception if the
   *  specified structure could not be found. Note that if the top element 
   *  on the stack is a structure that satisfies the name and type criteria, 
   *  select returns false.
   *  \brief Push the specified structure.
   *  \param type Requested structure type or "" to ignore the type.
   *  \param name Requested structure name or "" to ignore the name.
   *  \exception std::runtime_error is thrown if the specified structure was
   *             not found in the top stack element.
   *  \return True if the selected structure was pushed onto the stack.
   */
  bool select(const std::string& type, const std::string& name="");
  
private:
  std::unique_ptr<jsonStack> stack;
};

//======================================  inline methods
inline bool
jsonStream::empty(void) const {
  return (!stack || stack->empty());
}

inline bool
jsonStream::isArray(void) const {
  return (!empty() && stack->isArray());
}

#endif // !defined(JSONSTREAM_HH)
