/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef DATAMSG_HH
#define DATAMSG_HH

#include "lmsg/Message.hh"
#include "lmsg/TransInput.hh"
#include "lmsg/TransOutput.hh"

namespace lmsg {

   /**  Template for messages with a single data item. The data item may 
     *  be any defined class with the following properties.
     *  <ul>
     *    <li> A copy constructor</li>
     *    <li> An \c operator>>(TransInput&, T&) defined either intrinsically 
     *  in the \c TransInput class or as an external function.</li>
     *    <li> An \c operator<<(TransOutput&,const T&) defined either 
     *  intrinsically in the TransOutput class or as an external function.</li>
     *  </ul>
     *  @memo Data message template.
     *  @author John Zweizig
     *  @version $Revision$; Modified $Date$
     *  @ingroup IO_lmsg
     */
   template<class T, unsigned int MsgID>
   class DataMsg : public Message {
   public:

      /**  Construct a message.
        *  @memo Default constructor.
        */
      DataMsg(void) {}

      /**  Construct a message identical to an existing message.
        *  @memo Copy constructor.
        */
      DataMsg(const DataMsg<T,MsgID>& x) : mData(x.mData) {}

      /**  Construct a message and initialize data to a specified value.
        *  @memo Data constructor.
        */
      explicit DataMsg(const T& x) : mData(x) {}

      /**  Destroy a message.
        *  @memo Default destructor.
        */
      virtual ~DataMsg(void) {}

      /**  Get the message text length.
        *  @memo Text length.
        *  @return Always zero for dataless messages.
        */
      size_type size(void) const;

      /**  Write the message data to translation stream.
        *  @memo Get message data.
        */
      void getData(TransOutput& out) const;

      /**  Get the message type.
        *  @memo Message type.
        *  @return As specified by template argument.
        */
      MsgHeader::MsgType getType(void) const {return MsgID;}

      /**  Read data from a translation stream into the message.
        *  @memo Set message data.
        */
      void setData(TransInput& in);

      /**  Get a reference to the data portion of the message.
        *  @brief Reference message data
        *  @return Reference to data portion of message.
        */
      T& refData(void) {return mData;}

      /**  Get a reference to the data portion of the message.
        *  @brief Reference constant message data
        *  @return Constant reference to data portion of message.
        */
      const T& refData(void) const {return mData;}

   private:
      T mData;
   };

   //==================================  DataMsg method templates.
   template<class T, unsigned int MsgID>
   void
   DataMsg<T,MsgID>::getData(TransOutput& out) const {
      out << mData;
   }

   template<class T, unsigned int MsgID>
   void
   DataMsg<T,MsgID>::setData(TransInput& in) {
      in >> mData;
   }

   template<class T, unsigned int MsgID>
   size_type 
   DataMsg<T, MsgID>::size(void) const {
      TransOutput trout;
      getData(trout);
      return trout.getNBytes();
   }

}  // namespace lmsg

#endif   // DATAMSG_HH
