/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef LMSG_TRANSINPUT_HH
#define LMSG_TRANSINPUT_HH

#include <stdexcept>
#include <string>
#include "lmsg/MsgTypes.hh"
#include "lmsg/Grinder.hh"

class Time;

namespace lmsg {

    class Buffer;

    /**  Message input stream with translation capability. The TransInput 
      *  class is effectively an input stream that is intelligent enough 
      *  to translate between the big-endian external format and an internal
      *  format appropriate to the execution time machine.
      *
      *  @author John Zweizig
      *  @version 1.1; Modified August 20, 2001
      *  @ingroup IO_lmsg
      */
    class TransInput {
    public:
	/**  Construct an input stream and attach a full message buffer to it.
	 *  @memo data constructor.
	 *  @param b Message buffer to be read by the input stream.
	 */
	TransInput(const Buffer& b);

	/**  Default input stream constructor.
	 */
	TransInput(void);

	/**  Align data to appropriate boundary.
	 */
	void align(size_t bdry);

	/**  Test whether the specified number of bytes remain in the 
	  *  input buffer.
	  */
	bool available(size_t N) const;

	/**  Test whether there is any more data in the input buffer.
	 *  @memo Test whether buffer is empty;
	 *  @return true if the buffer is empty.
	 */
	bool empty(void) const;

	/** Get the remaining length of data in the translation buffer.
	 *  \brief Number of byte left in buffer.
	 *  \return Number of bytes.
	 */
	size_t getNBytes(void) const;

	/**  Template to read and translate a character array.
	 *  @memo read a character array.
	 *  @return number of characters read.
	 *  @param x Pointer to a character array to receive the data.
	 *  @param N Number of characters to be read.
	 */
	template<class T> 
	size_t read(T x[], size_t N=1);
 
	/** Stream input operator template;
	 */
	template <class T> 
	TransInput& operator>>(T& x);

    private:
	const char* mData;
	size_t      mIndex;
	size_t      mLength;
    };  //  class lmsg::TransInput

    //======================================  Specific templates

    /**  Read a character string of length N.
      *  @brief Read a character string
      *  @param x Character array to receive string.
      *  @param N Number of characters to read.
      *  @return Number of characters read.
      */
    template<>
    size_t TransInput::read(char x[], size_t N);

    /**  Read a dComplex array of length N.
      *  @brief Read a dComplex array
      *  @param x dComplex array to receive string.
      *  @param N Number of dComplex to read.
      *  @return Number of dComplex read.
      */
    template<>
    size_t TransInput::read(dComplex x[], size_t N);

    /**  Read a fComplex array of length N.
      *  @brief Read a fComplex array
      *  @param x fComplex array to receive string.
      *  @param N Number of fComplex to read.
      *  @return Number of fComplex read.
      */
    template<>
    size_t TransInput::read(fComplex x[], size_t N);

    /**  Read N message addresses.
      *  @brief Read MsgAddr's.
      *  @param x MsgAddr array to receive data.
      *  @param N Number of items to read.
      *  @return Number of data items read.
      */
    template<>
    size_t TransInput::read(MsgAddr x[], size_t N);

    /**  Read N GPS times.
      *  @brief Read GPS Time's.
      *  @param x Time array to receive data.
      *  @param N Number of items to read.
      *  @return Number of data items read.
      */
    template<>
    size_t TransInput::read(Time x[], size_t N);

    /**  Read N byte-count strings.
      *  @brief Read byte-count strings.
      *  @param x String array to receive data.
      *  @param N Number of items to read.
      *  @return Number of data items read.
      */
    template<>
    size_t TransInput::read(std::string x[], size_t N);

    //======================================  Inline methods
    template<class T>
    inline TransInput&
    TransInput::operator>> (T& x) {
	if (!read(&x, 1)) throw std::runtime_error("End-Of-Data");
	return *this;
    }

    inline bool 
    TransInput::available(size_t N) const {
	return mIndex + N <= mLength;
    }
  
    inline bool
    TransInput::empty(void) const {
	return mIndex >= mLength;
    }

    inline size_t
    TransInput::getNBytes(void) const {
	return mLength - mIndex;
    }

    template<class T> 
    inline size_t 
    TransInput::read(T x[], size_t N) {
	align(sizeof(T));
	if (empty()) return 0;
	if (!available(N*sizeof(T))) N = (mLength - mIndex)/sizeof(T);
	export_format_grinder.SwapIn(mData+mIndex, x, N);
	mIndex += N*sizeof(T);
	return N;
    }

}    // namespace lmsg

#endif  //  LMSG_TRANSINPUT_HH
