/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef LMSG_TRANSPORTMSG_HH
#define LMSG_TRANSPORTMSG_HH

#include "lmsg/MsgTypes.hh"

namespace lmsg {
   class Buffer;
   class BufferPool;
   class Message;
   class MsgAddr;
   class MsgHeader;
   class Socket;

   /**  TransportMsg provides an API for transporting messages with the lmsg
     *  protocol. This API formats mesages with the proper headers and sends
     *  or receives the messages using the lmsg::Socket UDP/IP message API.
     *  @memo Message transport API.
     *  @author John Zweizig
     *  @version 1.1; Modified April 21, 2000.
     *  @ingroup IO_lmsg
     */
   class TransportMsg {
   public:

      /**  Construct a message transporter.
       *  @memo TransportMsg constructor.
       */
      TransportMsg(void);

      /**  Close any open socket and destroy the message transporter.
       *  @memo TransportMsg destructor.
       */
      ~TransportMsg(void);

      /**  Get the address of the current socket.
       *  @memo Get address.
       *  @return Address to which the socket is bound.
       */
      virtual MsgAddr getAddr(void) const;

      /**  Get the reply address.
       */
      virtual MsgAddr getReplyAddr(void) const;

      /** Test whether transport is marked as being a asynchronous.
        *  \brief Test for asynchronous operation.
        *  \return true if asynchronous.
        */
      virtual bool is_async(void) const;

      /** Test whether transport is marked as being a client.
       *  \brief Test if a client.
       *  \return true if client instance.
       */
      virtual bool isClient(void) const;
      
      /**  Test whether transport is connected to a peer port.
       *  @memo Test for connected socket.
       *  @return True if socket is connected.
       */
      virtual bool    isConnected(void) const = 0;

      /**  Test whether transport has an open socket.
       *  @memo test for open socket.
       *  @return True if socket is open.
       */
      virtual bool    isOpen(void) const;

      /**  Get the current debugging level.
       *  @memo Get debug level.
       *  @return Current debugging level.
       */
      virtual index_type getDebug(void) const;

      /**  Open a socket with the specified option flags. An address to which 
       *  the socket will be bound may be specified if desired. If no address
       *  is specified, the socket will be bound to an unused IP address.
       *  @memo Open a socket.
       *  @return Lmsg error codes (OK or SysError).
       *  @param flags Bit mask specifying socket handling options.
       *  @param addr  Optional IP address to which the socket is to be bound.
       */
      virtual error_type open(mask_type flags, const MsgAddr* addr=0) = 0;

      /**  Return the socket to the appropriate pool.
       *  @memo Close the socket.
       */
      virtual void close(void) = 0;

      /**  Make a connection with the specified IP address. Once a connection 
       *  has been made messages may only be sent to or received from the 
       *  specified partner socket.
       *  @memo Connect to an IP partner address.
       *  @return Lmsg error codes (OK or SysError).
       *  @param  addr Address of partner to which the connection will be made.
       */
      virtual error_type connect(const MsgAddr& addr) = 0;

      /**  Disconnect from a partner.
       *  @memo Disconnect from a partner.
       *  @return Lmsg error codes (OK or SysError).
       */
      virtual error_type disconnect(void) = 0;

      /**  A message is received. The header of the message is copied to 'from' 
       *  and the message text (up to the specified maximum length) is copied 
       *  to 'b'. The buffer pointer may be coded as zero if the maximum 
       *  length is zero. receive returns with a TimeOut error code if no 
       *  message is received within 'timeout' seconds. If 'timeout' < 0,  
       *  receive will not return until either a message is received, or the 
       *  the wait is interrupted by a caught signal.
       *  @memo Receive a message.
       *  @return Lmsg error codes (OK, TimeOut, Continue, SysError).
       *  @param from    Received message header.
       *  @param b       Pointer to the buffer to receive the message text;
       *  @param maxlen  Length of buffer 'b'.
       *  @param timeout Maximum time in seconds to wait for a message.
       */
      virtual error_type receive(MsgHeader& from, char* b, size_type maxlen, 
				 wtime_type timeout=-1.0) = 0;

      /**  A message is received. The header of the message is copied to 'from' 
       *  and the message text (up to the specified maximum length) is copied 
       *  to 'msg'. receive returns with a TimeOut error code if no  message 
       *  is received within 'timeout' seconds. If 'timeout' < 0, receive will 
       *  not return until either a message is received, or the the wait is 
       *  interrupted by a caught signal.
       *  @memo Receive a message.
       *  @return Lmsg error codes (OK, TimeOut, Continue, SysError).
       *  @param from    Received message header.
       *  @param msg     Message buffer to receive the message text;
       *  @param timeout Maximum time in seconds to wait for a message.
       */
      virtual error_type receive(MsgHeader& from, Message& msg,
				 wtime_type timeout=-1.0) = 0;

      /**  A message is received. The messge header and text are returned in a 
       *  Buffer, the address of which is stored in 'Data'. The Buffer must
       *  be returned to the appropriate buffer pool with Data->Return()
       *  once the buffer is no longer needed. receive returns with a TimeOut 
       *  error code if no message is received within 'timeout' seconds. If 
       *  'timeout' < 0, receive will not return until either a message is 
       *  received, or it is interrupted by a caught signal.
       *  @memo Receive a message.
       *  @return Lmsg error codes (OK, TimeOut, Continue, SysError).
       *  @param Data Pointer to receive the filled buffer address.
       *  @param timeout Maximum time in seconds to wait for a message.
       */
      virtual error_type receive(Buffer** Data, wtime_type timeout=-1.0) = 0;

      /**  A message is sent. The header of the message to be sent is supplied
       *  by 'to' which must contain the destination address, the message 
       *  type and the length of the message text. The message text is 
       *   contained in 'b'.
       *  @memo Send a message.
       *  @return Lmsg error codes (OK, Continue, SysError).
       *  @param to Message header.
       *  @param b  Pointer to the buffer containing the message text;
       */
      virtual error_type send(const MsgHeader& to, const char* b) = 0;

      /**  A message is sent. The header of the message to be sent is supplied
       *  by 'to' which must contain the destination address and the 
       *  transaction and block IDs. The message text is contained in 'b'.
       *  @memo Send a message.
       *  @return Lmsg error codes (OK, Continue, SysError).
       *  @param to Message header.
       *  @param b  Message to be sent.
       */
      virtual error_type send(const MsgHeader& to, const Message& b) = 0;

      /**  Wait for a message or a timeout.
       *  @memo Wait for a message.
       *  @return Lmsg error codes (OK, Continue, SysError).
       *  @param Timeout Timeout time.
       */
      virtual error_type waitMsg(wtime_type Timeout) = 0;

      /**  Wait for a message in one or more transports or a timeout.
       *  @memo Wait for a message.
       *  @return Lmsg error codes (OK, Continue, SysError, Timeout, NotOpen).
       *  @param Timeout Timeout time.
       *  @param N       Number of transports to wait on.
       *  @param list    List of transports to wait on.
       */
      static error_type waitAny(wtime_type Timeout, int N,
				const TransportMsg* list[]);

      /**  Delete a socket. This method is to be used only for sockets that 
       *  are broken in some way e.g. unexpected messages have been received
       *  over the buffer.
       */
      virtual void Purge(void) = 0;

      /**  Specify the pool from which buffers will be allocated to hold sent 
       *  or received messages.
       *  @memo Specify buffer pool.
       *  @param pool Point to the buffer pool.
       */
      virtual void setBufferPool(BufferPool* pool);

      /**  Set the debugging level.
       *  @brief set the debug level
       *  @param level Debug level.
       */
      virtual void setDebug(index_type level);

      /**  Set the reply address.
       *	 @brief Set reply address.
       *  @param ad Address to which a reply message will be sent.
       */
      virtual void setReplyAddr(const MsgAddr& ad);

   protected:
      virtual error_type send(Buffer* b) = 0;

   protected:
      mask_type   mFlags;
      Socket*     mSocket;
      BufferPool* mBuffPool;
      index_type  mDebug;
      MsgAddr     mReplyAddr;
   };

} // namespace lmsg

//======================================  Inline methods
inline bool
lmsg::TransportMsg::is_async(void) const {
   return (mFlags & o_async) != 0;
}

inline bool
lmsg::TransportMsg::isClient(void) const {
   return (mFlags & o_client) != 0;
}

inline bool
lmsg::TransportMsg::isOpen(void) const {
   return (mSocket != 0);
}

inline lmsg::index_type
lmsg::TransportMsg::getDebug(void) const {
   return mDebug;
}

inline lmsg::MsgAddr
lmsg::TransportMsg::getReplyAddr(void) const {
   return mReplyAddr;
}

#endif // LMSG_SOCKET_HH
