/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef RESAMPLER_HH
#define RESAMPLER_HH

#include "Pipe.hh"
#include "autopipe.hh"
#include "FIRdft.hh"
#include "Time.hh"

/**  Resample a data stream with FIR anti-aliasing filter. The resampled data 
  *  are constrained to have the same phase as the original data and the time
  *  delay will be an integer number of output samples. At present this class
  *  performs down-sampling only. 
  *  \brief resample data
  *  \author John Zweizig
  *  \version 0.1
  */
class resampler : public Pipe {
public:
   /**  Default resampler constructor.
     *  \brief Default constructor.
     */
   resampler(void);

   /**  Initializing resampler constructor.
     *  \brief Default constructor.
     *  \param to_rate Sample rate of resampled series.
     */
   resampler(double to_rate);

#if __cplusplus > 201100L
   /**  Construct an exact copy of the specified resampler instance.
     *  \brief copy constructor
     *  \param x resampler to be copied.
     */
   resampler(const resampler& x) = default;
#endif
   
   /**  Resampler destructor.
     *  \brief Destructor.
     */
   virtual ~resampler(void);

   /**  Clone a resampler instance.
     *  \brief Clone clone this resampler.
     *  \returns Pointer to resampler clone.
     */
   resampler* clone(void) const;
   
   /**  Resample the input time series. 
     *  Synonymous with operator() method.
     *  \brief Resample an 
     *  \param in The TSeries being acted on.
     *  \return Resample time series.
     */
   virtual TSeries apply(const TSeries& in);

   /**  Validate an input time series. The input data are checked for 
     *  continuity and consistency with previously processed data. 
     *  \exception A std::runtime_error exception is thrown on invalid 
     *  input. For an exception-less version, see isDataValid(). 
     *  \brief Validate filter input. 
     *  \param in time series to be validated.
     */
   virtual void dataCheck(const TSeries& in) const;

   /** Reset the filter history.
     * \brief Reset the filter history.
     */
   virtual void reset(void);

   /**  Test whether the filter is being used. A filter is defined as being 
     *  in use if the filter has been applied to one or more time series 
     *  since the filter was constructed or reset. The filter state generally 
     *  reflects the application of the filters but need not.
     *  \brief Test if the filter is being used.
     *  \return true if the filter is in use.
     */
   virtual bool inUse(void) const;

   /**  Get the time of the first sample processed after the most recent
     *  reset.
     *  \brief Start time.
     *  \return Output keychain.
     */
   virtual Time getStartTime(void) const;

   /**  Get the time of the expected next sample to be processed.
     *  \brief Get the current time.
     *  \return Time of next expected sample.
     */
   virtual Time getCurrentTime(void) const;

   /**  Calculate the time delay imposed by this filter.
     *  \brief delay time.
     *  \return Delay time.
     */
   Interval getTimeDelay(void) const;

   /**  Set antialiasing filter parameters. The edge argument specifies the 
     *  width of the antialiasing filter transition zone in units of the input
     *  signal Nyquist frequency. The atten_dB argument specifies the 
     *  attenuation in dB.
     *  \brief Set the antialiasing filter parameters.
     *  \param edge Edge width in nyquist units (0.05). 
     */
   void set_AA_parameters(double edge, double atten_dB);

   /** Set the target sample rate. The interpolation and decimation stages 
    *  will be set up when the input sample rat is determined.
    *  \brief set the sample rate.
    *  \param to_rate Target sample rate. 
    */
   void set_sample_rate(double to_rate);

   /**  Set the FIR delay calculation mode. The default is to create a 
     *  zero-phase-shift filter, but the ode may also be set for the other 
     *  fir_mode options
     *  \brief Set the anti-aliasing delay mode.
     *  \param fmode FIR filter delay mode (as int(FIRdft::fir_mode))
     */
   void setMode(int fmode);
   
   /**  Set the interpolation and antialiasing filters.
    */
   void setup(void);

private:
   double    sample_rate;
   double    edgewid;          ///< fractional width of transition zone [0.02]
   double    attenuation;      ///< attenuation in dB [60dB]
   int       firmode;          ///< Filter delay mode [fm_zero_phase] 
   auto_pipe interpolate;
   auto_pipe antiAlias;
   int       defacto;
   Interval  inStep;
   TSeries   aabuffer;
   Time      startTime;
   Time      currentTime;
};

//======================================  resamplee inline methods
inline Time
resampler::getCurrentTime(void) const {
   return currentTime;
}

inline Time
resampler::getStartTime(void) const {
   return startTime;
}

inline bool
resampler::inUse(void) const {
   return startTime.ne_0();
}
#endif // !defined(RESAMPLER_HH)
